open Tk;;

(* Link a scroll bar to a text widget *)
let scroll_link sb txt =
  Text.configure txt [YScrollCommand (Scrollbar.set sb)];
  Scrollbar.configure sb [ScrollCommand (Text.yview txt)];;

(* The kill buffer *)
let make_kill_buffer text cb =
  let kill_ring = ref [] in
  let add_to_kill_ring s =
    kill_ring := s :: !kill_ring in
  let get_killed_text () =
    match !kill_ring with
    | [] -> ""
    | s :: l -> s in

  (* Note: for the text widgets, the insertion cursor is 
     not TextIndex (Insert, []),
     but TextIndex (Mark  "insert", []) *) 
  let insertMark = TextIndex (Mark "insert", []) in
  let eol_insertMark =
    TextIndex (Mark "insert", [LineEnd]) in

  let kill () =
    let s = Text.get text insertMark eol_insertMark in
    add_to_kill_ring (s ^ "\n");
    prerr_endline ("Killed: " ^ s);
    Text.delete text insertMark eol_insertMark in

  let yank () =
    let s = get_killed_text () in
    Text.insert text insertMark s [];
    prerr_endline ("Yanked: " ^ s) in

  let yank_more () =
    let ring = !kill_ring in
    let more = ref ring in
    let rec get_killed_more () =
      match !more with
      | [] -> more := ring; get_killed_more ()
      | s :: l -> more := l; s in
    let insert_killed_more () =
      let s = get_killed_more () in
      prerr_endline ("Yanked more: " ^ s);
      Text.insert text insertMark s [] in
    insert_killed_more ();
    bind text [[Alt], KeyPressDetail "y"]
      (BindSet ([], fun _ -> insert_killed_more ())) in

  bind text [[Control], KeyPressDetail "y"]
    (BindSet ([], fun _ -> yank ()));

  bind text [[Alt], KeyPressDetail "y"]
    (BindSet ([], fun _ -> yank_more ()));

  bind text [[Control], KeyPressDetail "k"]
    (BindSet ([], fun _ -> kill ()));

  bind text [[Control], KeyPressDetail "c"]
    (BindSet ([],
      fun _ ->
       let s = Text.get text
         (TextIndex (LineChar (0,0), []))
         (TextIndex (End, [])) in
       cb s));;

let make_text_widget top cbtop =
  let f = Frame.create top [] in
  let text = Text.create f [] in
  let scrollbar = Scrollbar.create f [] in
  scroll_link scrollbar text;
  pack [text; scrollbar][Side Side_Left; Fill Fill_X; Fill Fill_Y];
  pack [f][];
  let cb s =
    destroy f;
    cbtop s in
  make_kill_buffer text cb;;

let top = openTk ();;
make_text_widget top
 (fun s ->
    closeTk ();
    print_string s;
    print_newline ());;

mainLoop ();;
